from __future__ import annotations as __anotations
from typing import Callable as __C, Any as __Any
from ..__Util import Util as ut
import numpy as np
import scipy.cluster.hierarchy as Hierarchy
import scipy.spatial.distance as Distance
import matplotlib.pyplot as plt
from .__tdscurve import TDScurve
from .__tdscontainer import TDScontainer


class Hcluster:
    def __init__(self, tds_curves: TDScontainer, distance_func: __C = TDScurve.distance,
                 clustering_method: str = "ward", metric: str = "euclidean", optimal_ordering: bool = False,
                 threshold: __Any = None, criterion: str = "distance", depth: int = 2,
                 R: __Any | None = None, monocrit: __Any | None = None,):

        # print status
        ut.log("starting ", ("hierarchical clustering", ut.YELLOW), "...")

        self.tds_curves: TDScontainer = tds_curves

        # print status
        ut.log("calculating...")

        # distance function
        if isinstance(distance_func, str):
            distance_func = getattr(TDScurve, distance_func)

        # distance matrix
        self.distance_matrix: np.ndarray = Distance.squareform(
            [[distance_func(tds_curves[i], tds_curves[j]) for i in range(len(tds_curves))]
             for j in range(len(tds_curves))]
        )

        # linkage and clusters
        self.linkage: np.ndarray = Hierarchy.linkage(
            self.distance_matrix,
            method=clustering_method,
            metric=metric,
            optimal_ordering=optimal_ordering)
        self.clusters: np.ndarray = Hierarchy.fcluster(
            self.linkage,
            t=threshold or 0.7 * max(self.linkage[:, 2]),
            criterion=criterion,
            depth=depth,
            R=R,
            monocrit=monocrit)

    def draw(self, size: tuple[int] = (6, 4), font_family: str = "sans-serif",
             font_size: int = None, show: bool = True) -> Hcluster:
        if font_family:
            plt.rcParams["font.family"] = font_family
        if font_size:
            plt.rcParams["font.size"] = font_size

        ax = plt.figure(figsize=size).add_subplot()

        Hierarchy.dendrogram(self.linkage, ax=ax, leaf_label_func=lambda i: self.tds_curves[i].name)

        plt.tight_layout()
        show and plt.show()
        return self
