from __future__ import annotations as __anotations
import glob
import os
import matplotlib.pyplot as plt
from ..__Util import Util as ut
from .__tdscurve import TDScurve
from .__tdscontainer import TDScontainer
from .__pca import PCA
from .__hcluster import Hcluster


def loadFile(file_path: str, resolution: int = 1000, print_status: bool = True) -> TDScontainer:
    """This function generates a TDScontainer object from a single file.
    The file must be in the TDsampler output format.

    # Args
    * file_path  (str) : File path including the extension.
    * resolution (int) : Number of discretization of the whole time (start to stop). Defaults to 1000.

    # Returns
    * TDScontainer: A list-like object that contains multiple TDScurve objects.
    """

    file_path: str = file_path.replace("\\", "/")
    print_status and ut.log("loading ", (F'"{file_path}"', ut.GREEN), "...")

    with open(file_path, "r", encoding="UTF-8", newline="\n") as tdDataFile:
        trials: list[str] = tdDataFile.read().split("\n-\n")
        name = os.path.splitext(os.path.basename(file_path))[0]
        return TDScontainer(TDScurve.createFromCSV(trial, resolution=resolution).setName(name + (F'({i + 1})' if i else ""))
                            for i, trial in zip(range(len(trials)), trials))


def loadFolder(folder_path: str = ".", file_extension: str = ".csv",
               resolution: int = 1000, print_status: bool = True) -> TDScontainer:
    """This function generates a TDScontainer object from files in the specified folder(s).
    All files must be in the TDsampler output format.

    # Args
    * folder_path    (str) : Folder path (not the file path). You can use regular expressions. Defaults to ".".
    * file_extension (str) : File extension. Defaults to ".csv".
    * resolution     (int) : Number of discretization of the whole time (start to stop). Defaults to 1000.

    # Returns
    * TDScontainer: A list-like object that contains multiple TDScurve objects.
    """

    folder_path: str = folder_path.replace("\\", "/")
    print_status and ut.log("loading ", (F'"{folder_path}"', ut.GREEN), "...")

    files: list = glob.glob(F'{folder_path}/*{file_extension}', recursive=True)
    obj: TDScontainer = TDScontainer(curve for f in files for curve in loadFile(f, resolution=resolution, print_status=False))

    if not len(obj):
        raise Exception("Folder seems to be empty.")

    return obj


def repl(vars: dict):
    import code
    code.InteractiveConsole(vars).interact()


def show(*args, **kwargs) -> None:
    plt.show(*args, **kwargs)


# add some font parameters to matplotlib
plt.rcParams["font.sans-serif"][0:0] = ("Meiryo", "Yu Gothic", "MS Gothic", "Hiragino Sans")
