from __future__ import annotations as __anotations
from typing import Sequence as __S, Callable as __C
from ..__Util import Util as ut
import numpy as np
import scipy.stats as Stats
import matplotlib.pyplot as plt
from .__tdscurve import TDScurve
from .__tdscontainer import TDScontainer


# class for Principle Components Analysis (PCA)
class PCA:

    def __init__(self, tds_curves: TDScontainer, number_of_pcs: int = 2,
                 data_extractor: __C[[TDScurve], __S[__S[float]]] = lambda x: x.dominance_duration[:-1],
                 variables: __S[str] | dict[str, int] | __C[[TDScurve], __S[str] | dict[str, int]] = lambda x: x.attributes,
                 print_result: bool = True):

        # print status
        ut.log("starting ", ("principle components analysis", ut.YELLOW), "...")

        # variables function
        if isinstance(variables, dict):
            def variables_func(_): return variables
        elif hasattr(variables, "__iter__") and hasattr(variables, "__getitem__"):
            def variables_func(_): return dict(zip(variables, range(len(variables))))
        else:
            variables_func = variables

        # check the type of tds_curves and variables
        first_elem: TDScurve = tds_curves[0]
        if not all(isinstance(curve, TDScurve) and variables_func(curve) == variables_func(first_elem)
                   for curve in tds_curves):
            raise Exception("Different formats are mixed. Please review attribute words.")

        # variables from arguments
        self.tds_curves: TDScontainer = tds_curves
        self.number_of_pcs: int = number_of_pcs
        vars = variables_func(first_elem)
        self.variables: dict[str, int] = vars if isinstance(vars, dict) else dict(zip(vars, range(len(vars))))

        # print status
        ut.log("calculating...")

        # calculate normalized data and correlation coefficient
        raw_data: list = [data_extractor(curve) for curve in tds_curves]
        self.data: np.ndarray = Stats.zscore(raw_data).T
        self.corrcoef: np.ndarray = np.corrcoef(self.data)

        # eigenvalues, eigenvectors
        [w, v] = np.linalg.eig(self.corrcoef)
        index = np.argsort(w)[::-1]

        self.eigenvalues: np.ndarray = w[index]
        self.eigenvectors: np.ndarray = v.T[index]

        # pc scores and loadings for each item
        self.scores: np.ndarray = np.array([self.data.T @ self.eigenvectors[i] for i in range(number_of_pcs)])
        self.loadings: np.ndarray = np.array([np.identity(len(self.eigenvectors.T)) @ self.eigenvectors[i]
                                              for i in range(number_of_pcs)])

        # print analysis result
        if print_result:
            self.printValues()

    def rotateVarimax(self, iteration_limit: int = 10000) -> PCA:
        """
        varimax rotation
        """
        loadings: np.ndarray = self.loadings.T.copy()
        [n, p] = loadings.shape
        h: np.ndarray

        if p > 1:
            h = (loadings ** 2).sum(1) ** 0.5
            loadings = (loadings.T / h).T
        else:
            return loadings

        for _ in range(iteration_limit):
            # copy of the original matrix
            original: np.ndarray = loadings.copy()

            for i in range(0, p):
                for j in range(i + 1, p):

                    [x, y] = [loadings[:, i].copy(), loadings[:, j].copy()]
                    [u, v] = [x ** 2 - y ** 2, 2 * x * y]
                    [a, b, c, d] = np.array([u, v, u ** 2 - v ** 2, 2 * u * v]).sum(1)

                    # coefficients
                    num = d - 2 * a * b / n
                    den = c - (a ** 2 - b ** 2) / n

                    # angle
                    theta = np.arctan(num / den)
                    theta += np.pi * (num * theta < 0)
                    theta /= 4.0

                    # rotate
                    cos = np.cos(theta)
                    sin = np.sin(theta)
                    loadings[:, i] = x * cos + y * sin
                    loadings[:, j] = -x * sin + y * cos

            # floating point error
            err: float = ((original - loadings) ** 2).sum()
            if (np.isclose(err, 0)):
                loadings = (loadings.T * h).T
                break

        p = loadings @ np.linalg.inv(loadings.T @ loadings)
        self.scores = (self.scores.T @ self.loadings @ p).T
        self.loadings = loadings.T

        return self

    def rotatePromax(self, power: float = 4.0, iteration_limit: int = 10000) -> PCA:
        """
            promax rotation
        """

        # varimax rotation
        self.rotateVarimax(iteration_limit)

        # promax rotation
        loadings = self.loadings.T.copy()
        u = loadings.T / (loadings ** 2).sum(1) ** 0.5
        c = np.sign(loadings) * np.abs(u.T / np.max(np.abs(u), 1)) ** power
        q = np.linalg.inv(loadings.T @ loadings) @ loadings.T @ c
        d = np.diag(np.linalg.inv(q.T @ q).diagonal() ** 0.5)

        Lambda = q @ d
        loadings = loadings @ Lambda

        p = loadings @ np.linalg.inv(loadings.T @ loadings)
        self.scores = (self.scores.T @ self.loadings @ p).T
        self.loadings = loadings.T

        # factor correlation matrix
        corr: np.ndarray = np.linalg.inv(Lambda.T @ Lambda)

        return self

    def printValues(self) -> PCA:
        for i in range(self.number_of_pcs):
            ut.printc((F"Score of PC{i + 1}:", ut.YELLOW), F'\n{self.scores[i]}\n')

        ut.printc(("Eigenvalues:", ut.YELLOW), F'\n{self.eigenvalues}\n')
        ut.printc(("Each Contribution Rate (%):", ut.YELLOW), F'\n{self.eigenvalues / len(self.eigenvalues) * 100}\n')
        ut.printc(("Eigenvectors:", ut.YELLOW), F'\n{self.eigenvectors}\n')
        return self

    def draw(self, clusters: __S[int] = None, size: tuple[float] = (6, 4), font_family: str = "sans-serif", font_size: int = None,
             point_size: int = 20, show: bool = True, show_legend: bool = True, show_grid: bool = True,
             show_scatter: bool = True) -> PCA:

        if font_family:
            plt.rcParams["font.family"] = font_family
        if font_size:
            plt.rcParams["font.size"] = font_size

        for i in range(len(self.scores) - 1):
            for j in range(i + 1, len(self.scores)):
                [xlabel, ylabel] = [F'PC{i + 1}', F'PC{j + 1}']
                fig = plt.figure(figsize=size)
                fig.suptitle(F'{xlabel} and {ylabel}')
                ax: plt.Axes = fig.add_subplot(xlabel=F'{xlabel} ({int(self.eigenvalues[i] / len(self.eigenvalues) * 100)}%)',
                                               ylabel=F'{ylabel} ({int(self.eigenvalues[j] / len(self.eigenvalues) * 100)}%)')

                if clusters is not None:
                    c: int = 0
                    color = plt.rcParams["axes.prop_cycle"].by_key()["color"]
                    no_cluster = True
                    for k in range(max(clusters)):
                        xp = [self.scores[i][m] for m in range(len(clusters)) if clusters[m] - 1 == k]
                        yp = [self.scores[j][m] for m in range(len(clusters)) if clusters[m] - 1 == k]
                        in_cluster = len(xp) > 1
                        if in_cluster:
                            c += 1
                            marker = "D" if k >= 10 else "o"
                            ax.scatter(xp, yp, color=color[c % 10], label=F'cluster{c}', s=point_size, marker=marker)
                        elif no_cluster:
                            ax.scatter(xp, yp, color=color[0], label="not in cluster", s=point_size, marker="x")
                            no_cluster = False
                        else:
                            ax.scatter(xp, yp, color=color[0], s=point_size)
                elif show_scatter:
                    ax.scatter(self.scores[i], self.scores[j], s=point_size, marker="o")

                # plot loadings
                for key in self.variables:
                    k: int = self.variables[key]
                    linestyle: str = "dashed" if k >= 10 else "solid"
                    ax.plot([0, self.loadings[i][k]], [0, self.loadings[j][k]], label=key, linestyle=linestyle)

                show_grid and ax.grid()
                show_legend and ax.legend()

        plt.tight_layout()
        show and plt.show()
        return self
