from __future__ import annotations as __anotations
from typing import Iterator as __Iter, Sequence as __S, Callable as __C, Any as __A
import warnings
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.figure as figure
from ..__Util import Util as ut


class TDScurve:
    @classmethod
    def createFromCSV(cls, tab_separated_values: str, resolution: int = 1000) -> TDScurve:
        meta: dict[str, list] = {}
        time: dict[str, list[float]] = {}
        lines: list[str] = tab_separated_values.split("\n")

        for line in lines:
            params: list[str] = [s.strip() for s in line.split(":")[0].split("\t")]
            paramlen: int = len(params)
            attr: str = params[0]
            if attr == "" or attr in "!\"$%&'()*+,-./;<=>?@[\\]^_`{|}~":
                continue
            if paramlen > 1:
                if attr[0] == "#":
                    tag: str = attr[1:].upper()
                    if tag in {"DURATION", "DELAY"} and params[1]:
                        meta[tag] = [float(params[1])]
                    elif tag in {"SCORE", "SCORE_MAX", "SCORE_MIN", "COUNT", "YEAR"}:
                        meta[tag] = [int(params[1])] if params[1] else []
                    else:
                        meta[tag] = params[1:]
                else:
                    time[attr] = [float(param) for param in params[1:] if param]

            elif paramlen == 1:
                if attr[0] == "#":
                    meta[attr[1:].upper()] = []
                else:
                    time[attr] = []

        if "DURATION" not in meta:
            meta["DURATION"] = [max(max(time[key]) for key in time if time[key])]
            ut.log(("The duration data was not found, so it was automatically generated.", ut.MAGENTA))

        if "DELAY" not in meta:
            meta["DELAY"] = [min(min(time[key]) for key in time if time[key])]
            ut.log(("The delay data was not found, so it was automatically generated.", ut.MAGENTA))

        duration: float = meta["DURATION"][0]
        data: np.ndarray = np.zeros((len(time) + 1, resolution), float)
        attribute_num: int = 0
        for key in sorted(time):
            sample: list = time[key]
            for i in range(0, len(sample), 2):
                timebegin = sample[i]
                timeend = sample[i + 1]
                data[attribute_num, int(timebegin / duration * resolution): int(timeend / duration * resolution)] = 1.0

            time[key] = attribute_num
            attribute_num += 1

        data[-1] = 1.0 - data.sum(0)

        if not np.allclose(np.isclose(data[-1], 0) ^ np.isclose(data[-1], 1), True):
            raise Exception("Data may be corrupted.")

        return TDScurve(meta=meta, data=data, attributes=time)

    @classmethod
    def sum(cls, args: list[TDScurve]) -> TDScurve:
        first: TDScurve = args[0]
        new_meta: dict = copy.deepcopy(first.meta)
        new_data: np.ndarray = copy.deepcopy(first.data) * first.trials_count
        buff: np.ndarray = np.empty(new_data.shape, float)

        for elem in args[1:]:

            cls.__checkOperable(first, elem)
            for key in {*first.meta} & {*elem.meta}:
                new_meta[key] += elem.meta[key]

            np.multiply(elem.data, elem.trials_count, buff)
            new_data += buff

        new_data /= new_data.sum(0)

        name: str
        if "ASSESSOR" in first.meta:
            name = F'{first.meta["ASSESSOR"][0]} and {sum(a.trials_count for a in args) - 1} others'
        else:
            name = F'{sum(a.trials_count for a in args)} trials'
        return TDScurve(name=name, meta=new_meta, data=new_data, attributes=first.attributes).fix()

    @property
    def trials_count(self) -> int:
        return len(self.meta["DURATION"])

    @property
    def resolution(self) -> int:
        return self.data.shape[1]

    @property
    def delay(self) -> np.ndarray:
        return self.data[-1]

    @property
    def dominance_duration(self) -> np.ndarray:
        return self.data.sum(1) / self.resolution

    def __init__(self, name: str = None, meta: dict = None,
                 data: np.ndarray = None, attributes: dict = None):

        self.name: str = name or "No Name"
        self.meta: dict = dict(sorted(meta.items()))
        self.data: np.ndarray = data
        self.attributes: dict[str, int] = attributes

        if meta is None or data is None or attributes is None or self.trials_count != len(meta["DELAY"]):
            raise Exception("Data may be corrupted.")
        elif "DELAY" in attributes:
            warnings.warn("It is not recommended to use \"DELAY\" as an attribute word.", Warning)

    def __add__(self, other: TDScurve) -> TDScurve:
        return TDScurve.sum((self, other)) if other else self

    def __radd__(self, other: TDScurve) -> TDScurve:
        return other + self if other else other

    def __iadd__(self, other: TDScurve) -> TDScurve:
        self = self + other
        return self

    def __sub__(self, other: TDScurve) -> np.ndarray:
        self.__checkOperable(other)
        return self.data - other.data

    def __rsub__(self, other: TDScurve) -> np.ndarray:
        return -(self - other)

    def __matmul__(self, func: __C) -> __A:
        return func(self)

    def __eq__(self, other: TDScurve) -> bool:
        return self is other

    def __ne__(self, other: TDScurve) -> bool:
        return not (self == other)

    def __rshift__(self, func: __C) -> __A:
        return func(self.data)

    def __getitem__(self, key: str) -> np.ndarray:
        return self.data[self.attributes[str(key)]]

    def __call__(self, normalized_time: float = 0.0) -> np.ndarray:
        time: int = max(int(normalized_time * self.resolution) - 1, 0)
        dp: np.ndarray = self.data[:, time] if time else np.array([0.0] * len(self) + [1.0])
        return dp.copy()

    def __abs__(self) -> int:
        return self.trials_count

    def __len__(self) -> int:
        return len(self.attributes)

    def __iter__(self) -> __Iter[str]:
        return self.attributes.__iter__()

    def __repr__(self) -> str:
        return F'[TDScurve of {self.trials_count} trial{"" if self.trials_count <= 1 else "s"}]'

    def __bool__(self) -> bool:
        return True

    def __hash__(self) -> int:
        return id(self)

    def mapAttributes(self, array: str | __S) -> dict[str, __A]:
        attrs = sorted(self.attributes, key=lambda i: self.attributes[i])
        if isinstance(array, str):
            array = getattr(self, array)
        return dict(zip((*attrs, "DELAY"), array))

    def getAverageDelay(self) -> float:
        return np.mean(self.meta["DELAY"])

    def getAverageDuration(self) -> float:
        return np.mean(self.meta["DURATION"])

    def getMeta(self, key: str) -> __A:
        key: str = str(key).upper()
        meta: list = self.meta[key] if key in self.meta else []
        meta_len: int = len(meta)
        if not meta_len:
            warnings.warn(F'No metadata for "{key}"', Warning)
        elif meta_len > 1:
            warnings.warn(F'Metadata "{key}" has multiple values.', Warning)

        return meta[0] if meta_len else None

    def getMetas(self, key: str) -> list[__A]:
        key: str = str(key).upper()
        meta: list = self.meta[key] if key in self.meta else []
        if not len(meta):
            warnings.warn(F'No metadata for "{key}"', Warning)
        return meta

    def distance(self, other: TDScurve) -> float:
        self.__checkOperable(other)
        return np.mean((((self.data - other.data) ** 2).sum(0) / 2) ** 0.5)

    def distanceM(self, other: TDScurve) -> float:
        self.__checkOperable(other)
        return np.mean(((abs(self.data - other.data)).sum(0) / 2))

    # test ###########################

    def distanceTest1(self, other: TDScurve) -> float:
        self.__checkOperable(other)
        return 1 - np.prod(((self.data * other.data) ** (1 / self.resolution)).sum(0))

    def distanceTest2(self, other: TDScurve) -> float:
        self.__checkOperable(other)
        return 1 - np.mean((self.data * other.data).sum(0))

    def distanceTest3(self, other):
        def mat(x):
            z = (np.concatenate([np.array([x(0)]).T, x.data[:, :-1]], 1) @ x.data.T)
            z = z - np.diag(np.diag(z)) + np.identity(1 + x @ len)
            z = z / z.sum(1)
            return z.T
        return np.mean((((mat(self) - mat(other)) ** 2).sum(0) / 2) ** 0.5)

    # end test ###########################

    def ddDistance(self, other: TDScurve) -> float:
        self.__checkOperable(other)
        return (((self.dominance_duration - other.dominance_duration) ** 2).sum() / 2) ** 0.5

    def setName(self, name: str = "No Name") -> TDScurve:
        self.name = str(name)
        return self

    def setMeta(self, key: str, value: __A) -> TDScurve:
        self.meta[str(key).upper()] = [value]
        return self

    def convolve(self, weight: int = 10) -> TDScurve:
        edge: int = int(int(weight) / 2)
        rows: int = len(self.data)
        edged: np.ndarray = np.concatenate([np.tile(self.data[:, 0].reshape(rows, 1), edge), self.data,
                                            np.tile(self.data[:, -1].reshape(rows, 1), weight - edge - 1)], 1)
        self.data = np.array([np.convolve(row, [1 / weight] * weight, "valid") for row in edged])
        return self

    def fix(self) -> TDScurve:
        self.data /= self.data.sum(0)
        return self

    def resample(self, resolution: int, phase: float = 1.0) -> TDScurve:
        width: int = self.resolution // resolution
        self.data = self.data[:, range(min(int(width * phase), width - 1), self.resolution, width)]
        return self

    def draw(self, layout: list = None, scale: float = 1.0, size: tuple[float] = (6, 4), font_family: str = "sans-serif",
             font_size: int = None, show: bool = True, auto_adjust: bool = True, legend_location: int = 0,
             show_chance: bool = True, show_signif: bool = True, show_total: bool = False,
             show_delay: bool = False, show_average_delay: bool = False) -> TDScurve:

        if layout is None:
            layout = [[[*self.attributes]]]

        if font_family:
            plt.rcParams["font.family"] = font_family
        if font_size:
            plt.rcParams["font.size"] = font_size

        rows: int = len(layout)
        columns: int = max(len(row) for row in layout)
        axes: list[list[plt.Axes]] = [[None] * columns] * rows
        count: int = 1
        standard_ax: np.ndarray = np.concatenate([np.arange(0, self.resolution) / self.resolution * scale, [scale]])
        subplot_args: dict = {}

        fig: figure.Figure = plt.figure(figsize=size)
        fig.suptitle(self.name)

        if not auto_adjust:
            subplot_args["ylim"] = (0, 1.0)

        for i, row in enumerate(layout):
            for j, column in enumerate(row):
                axes[i][j] = fig.add_subplot(rows, columns, count, xlabel="Normalized Time", ylabel="Dominance Proportion",
                                             **subplot_args)
                count += 1

                # TDS curves
                attributes_count = 1
                for key in column:
                    style: str = "dashed" if attributes_count > 10 else "solid"
                    axes[i][j].plot(standard_ax, np.concatenate([[0.0], self[key]]), label=key, linestyle=style)
                    attributes_count += 1

                # graph parameters
                chance = 1 / len(self.attributes)
                signif = chance + 1.64485362695145 * (chance * (1 - chance) / self.trials_count) ** 0.5
                average_delay = np.mean(self.delay) * scale

                # plot graph parameters
                show_chance and axes[i][j].plot((0, scale),
                                                (chance,) * 2,
                                                label="Chance",
                                                linestyle="dotted",
                                                color="gray")
                show_signif and axes[i][j].plot((0, scale),
                                                (signif,) * 2,
                                                label="Signif.",
                                                linestyle="dashed",
                                                color="gray")
                show_total and axes[i][j].plot(standard_ax,
                                               np.concatenate([[0.0], sum(self[key] for key in column)]
                                                              if not show_delay else
                                                              [[1.0], sum(self[key] for key in column) + self.delay]),
                                               label="Total",
                                               linestyle="dashed")
                show_delay and axes[i][j].plot(standard_ax,
                                               np.concatenate([[1.0], self.delay]),
                                               label="Delay",
                                               linestyle="dashed")
                show_average_delay and axes[i][j].plot((average_delay, average_delay),
                                                       (0, show_delay or (max(max(self[key].max() for key in column), chance, signif)
                                                                          if not show_total else
                                                                          sum(self[key] for key in column).max())),
                                                       label="Average Delay",
                                                       linestyle="dashdot",
                                                       color="gray")

                legend_location is None or axes[i][j].legend(loc=legend_location)

        # show graph or not
        plt.tight_layout()
        show and plt.show()

        return self

    def save(self, destination: str = ".", file_name: str = None, file_extension: str = ".csv", delimiter: str = "\t",
             include_delay: bool = False) -> TDScurve:

        data: np.ndarray = self.data if include_delay else self.data[:-1]
        data = data.round(4)

        if not file_name:
            file_name = self.name

        with open(F'{destination}/{file_name}{file_extension}', "w", encoding="UTF-8", newline="\n") as f:
            attrs = sorted(self.attributes, key=lambda i: self.attributes[i])
            result = F'{delimiter.join(attrs)}{delimiter + "DELAY" if include_delay else ""}\n' + \
                "\n".join(delimiter.join(map(str, row)) for row in data.T) + \
                "\n"
            f.write(result)
        return self

    def __checkOperable(self, other: TDScurve) -> None:
        if self.attributes != other.attributes:
            raise Exception("Different formats are mixed. Please review attribute words.")

        if self.resolution != other.resolution:
            raise Exception("Data of different lengths are mixed. Please review \"resolution\" field.")

        if other is not None and not isinstance(other, TDScurve):
            raise TypeError("Objects of unsupported types are mixed.")

        if {*self.meta} ^ {*other.meta}:
            warnings.warn("Metadata of different types are mixed. Please review \"meta\" field.", Warning)
